#include <stdio.h>
#include <TMath.h>
#include <TGraph2D.h>
#include <TRandom2.h>
#include <TStyle.h>
#include <TCanvas.h>
#include <TF2.h>
#include <TH1.h>
#include <TVirtualFitter.h>
#include <TPolyLine3D.h>
#include <Math/Vector3D.h>
#include <TApplication.h>
#include <TThread.h>
#include "/opt/pyrame/cmdmod_helper.hpp"

using namespace ROOT::Math;

struct or_ws {
  TGraph2D *graph;
  int nbp;
  struct evloop *loop;
  double time;
};

void *runner(void *appli) {
  TApplication *app=(TApplication *)appli;
  //run the TApp
  app->Run();
  return NULL;
}

// define the parameteric line equation 
void line(double t, double *p, double &x, double &y, double &z) { 
   // a parameteric line is define from 6 parameters but 4 are independent
   // x0,y0,z0,z1,y1,z1 which are the coordinates of two points on the line
   // can choose z0 = 0 if line not parallel to x-y plane and z1 = 1; 
  x=p[0]+p[1]*t; 
  y=p[2]+p[3]*t;
  z=t; 
}

// calculate distance line-point 
double distance2(double x,double y,double z, double *p) { 
   // distance line point is D= | (xp-x0) cross  ux | 
   // where ux is direction of line and x0 is a point in the line (like t = 0) 
  XYZVector xp(x,y,z); 
  XYZVector x0(p[0],p[2],0.); 
  XYZVector x1(p[0]+p[1],p[2]+p[3],1.); 
  XYZVector u=(x1-x0).Unit(); 
  double d2=((xp-x0).Cross(u)).Mag2(); 
  return d2; 
}

// function to be minimized 
void SumDistance2(int &, double *, double & sum, double * par, int ) { 
  // the TGraph must be a global variable
  TGraph2D * gr = dynamic_cast<TGraph2D*>( (TVirtualFitter::GetFitter())->GetObjectFit() );
  assert(gr != 0);
  double * x=gr->GetX();
  double * y=gr->GetY();
  double * z=gr->GetZ();
  int npoints=gr->GetN();
  sum=0;
  for (int i=0;i<npoints;++i) { 
    double d=distance2(x[i],y[i],z[i],par); 
    sum+=d;
  }
}

double line3Dfit(void *workspace)
{
  struct or_ws *ws=(struct or_ws *)workspace;
  gStyle->SetOptStat(0);
  gStyle->SetOptFit();
   
  TVirtualFitter *min = TVirtualFitter::Fitter(0,4);
  min->SetObjectFit(ws->graph);
  min->SetFCN(SumDistance2);
  
  
  Double_t arglist[10];
  arglist[0] = 3;
  min->ExecuteCommand("SET PRINT",arglist,1);
  
  double pStart[4] = {1,1,1,1};
  min->SetParameter(0,"x0",pStart[0],0.01,0,0);
  min->SetParameter(1,"Ax",pStart[1],0.01,0,0);
  min->SetParameter(2,"y0",pStart[2],0.01,0,0);
  min->SetParameter(3,"Ay",pStart[3],0.01,0,0);
  
  arglist[0]=1000; // number of function calls 
  arglist[1]=0.001; // tolerance 
  min->ExecuteCommand("MIGRAD",arglist,2);
  
  // get fit parameters
  double parFit[4];
  for (int i = 0; i <4; ++i) 
    parFit[i] = min->GetParameter(i);

  //compute the angle
  double x1,y1,z1;
  double t1=0;
  line(t1,parFit,x1,y1,z1);
  printf("first coord %g,%g,%g\n",x1,y1,z1);
  double x2,y2,z2;
  double t2=1000;
  line(t2,parFit,x2,y2,z2);
  printf("second coord %g,%g,%g\n",x2,y2,z2);
  double dist=std::sqrt((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2));
  printf("dist=%g\n",dist);
  double angle=TMath::ATan(dist/1000)*180.0/Pi();
  if (y1-y2<0)
    angle=angle*-1.0;
  angle=fmod(angle,180);
  printf("angle=%d\n",(int)angle);

  return angle;

}

void reinit_func(void *workspace) {
  return;
}

void newblock_func(void *workspace,struct block *b) {
  struct or_ws *ws=(struct or_ws *)workspace;
  
  printf("new block with %d points\n",b->nb_events);
  ws->graph=new TGraph2D(b->nb_events);
  ws->nbp=0;
}

void endblock_func(void *workspace,struct block *b) {
  struct or_ws *ws=(struct or_ws *)workspace;
  struct cmd_result *pres;
  char *block_id;
  char *event_id;
  double angle;

  //calc angle
  angle=line3Dfit(ws);
  
  
  //publish angle
  pres=submod_new_block(b->id);
  if (pres->retcode==0) {
    printf("cant create new block : %s\n",pres->str);
    free_cmd_result(pres);
    return;
  }
  block_id=pres->str;
  free(pres);
  pres=submod_new_event(block_id);
  if (pres->retcode==0) {
    printf("cant create new event : %s\n",pres->str);
    free(block_id);
    free_cmd_result(pres);
    return;
  }
  event_id=pres->str;
  free(pres);
  submod_set_field_event(event_id,"time","%f",ws->time);
  submod_set_field_event(event_id,"angle","%f",angle);
  submod_finalize_block(block_id);
  free(block_id);
  free(event_id);
}



void treatevt_func(void *workspace,struct event *e) {
  struct or_ws *ws=(struct or_ws *)workspace;

  //fill the histo with the energy
  Double_t x,y,z;
  x=(Double_t)atof(get_event_field(e,"x"));
  y=(Double_t)atof(get_event_field(e,"y"));
  z=(Double_t)atof(get_event_field(e,"z"));
  ws->time=atof(get_event_field(e,"time"));
  printf("adding point %d : %g %g %g\n",ws->nbp,x,y,z);
  ws->graph->SetPoint(ws->nbp,x,y,z);
  ws->nbp++;
}

extern "C" void*init(int,char **);
void *init(int argc, char **argv) {
  if (argc<1) {
    printf("usage : -a ds_name\n");
    submod_exit(1);
  }

  //initialize
  struct or_ws *ws=(struct or_ws *)malloc(sizeof(struct or_ws));
  memset(ws,0,sizeof(struct or_ws));
  TApplication *ecalApp=new TApplication("App",&argc,argv);
  struct cmd_result *pres;
  pthread_t *app_t;
  
  printf("init angle acq\n");
  pres=submod_init_dispatcher(10000,"angles","!time||angle","undef");
  if (pres->retcode==0) {
    printf("unable to init dispatcher : %s\n",pres->str);
    submod_exit(1);
  }
  submod_setmyport(atoi(pres->str));
  free_cmd_result(pres);

  ws->loop=new_event_loop(argv[0],ws,treatevt_func,newblock_func,endblock_func,reinit_func,FIRST);
  run_loop(ws->loop);

  //launch the TApplication
  app_t=(pthread_t*)malloc(sizeof(pthread_t));
  if (pthread_create(app_t,NULL,runner,(void *)ecalApp)<0) {
    perror("pthread_create");
    exit(1);
  }
  pthread_detach(*app_t);

  //never reached
  return ws;
}

