//
// Copyright 2012-2014 Frédéric Magniette, Miguel Rubio-Roy, LLR
//

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <pyrame.h>

//Skiroc specific parameters

#define NB_CHANS 64
#define NB_CHIPS 4
#define NB_SCA 16

#define loss_code 0xfffa

//other parameters

#define BACKLOG 100
#define BUFFER_SIZE 1024*1024 //in 16 bits words
#define INPUT_FILE 2
#define SHARED_BUFFER_SIZE 8192


struct dif_stream {
  int debug;
  //data reading
  unsigned short *buffer;
  unsigned short * data_start;
  int data_size; //in 16 bits words
  //system
  int input_mode;
  //files reading
  int ddid;
  char *filename;
  char *outputname;
  FILE * ftest;
  //geometry
  int nb_chips;
  int nb_chans;
  //control
  unsigned int current_spil;
  unsigned int current_chip;
  //statistics
  int nb_events;
  int nb_hits;
  int nb_loss;
  int nb_corrupt;
  int nb_gain_corrupt;
  int nb_plain;
  int nb_bpn;
  int nb_fb_corrupt;
  int nb_miss_ct;
  int nb_readout;
} dif_stream;

//initialize a new dif_stream structure
struct dif_stream * new_dif_stream(int nb_chips,int nb_chans) {
  struct dif_stream *result;
  result=malloc(sizeof(struct dif_stream));
  memset(result,0,sizeof(struct dif_stream));
  result->buffer=malloc(BUFFER_SIZE*sizeof(unsigned short));
  result->data_start=result->buffer;
  result->nb_chips=nb_chips;
  result->nb_chans=nb_chans;
  //initialize statistics
  result->nb_loss=0;
  result->nb_events=0;
  result->nb_hits=0;
  result->nb_corrupt=0;
  result->nb_gain_corrupt=0;
  result->nb_plain=0;
  result->nb_bpn=0;
  result->nb_fb_corrupt=0;
  result->nb_miss_ct=0;
  result->nb_readout=0;
  return result;
}

//this function abstract the reading mode and acquire a bunch of data
//return 0 if read is ok
//return -1 else
int acquire_next_chunk(struct dif_stream *ds) {
  int read_size;
  //int total;

  //if no room left, recopy the data at the beginning of the buffer
  if (ds->data_start-ds->buffer>BUFFER_SIZE-(SHARED_BUFFER_SIZE*3)) {
    memcpy(ds->data_start,ds->buffer,ds->data_size);
    ds->data_start=ds->buffer;
  }

  //reading the file
  if (ds->input_mode==INPUT_FILE) {
    //read a bunch of data 
    read_size=read(ds->ddid,ds->data_start+ds->data_size,SHARED_BUFFER_SIZE);
    if (read_size!=SHARED_BUFFER_SIZE) { 
      if (ds->debug) {
	printf("read failed %d\n",read_size);
	perror("read");
      }
      //end of file
      return -1;
    } else {
      //good reading
      ds->data_size+=SHARED_BUFFER_SIZE/2; //in words
      return 0;
    }
    //never reached
  }

  if (ds->debug)
    printf("impossible mode : %d\n",ds->input_mode);
  return -1;
}

void display_sample(struct dif_stream *ds) {
  int j;
  for (j=0;j<10;j++) {
    printf("%x : %x\t",j,ds->data_start[j]);
    if (j==4) printf("\n");
  }
  printf("\n");
}


//search for the 0xffff spill stop pattern in the data stream
//manage the error pattern
//return 0 if success
//return -1 if failure
int search_next_spil_stop(struct dif_stream *ds) {
  int i;
  int ignore_next=0;
  int res;
  if (ds->debug)
    printf("search_next_spil_stop\n");

  while(1) { //main loop

    //acquire data if necessary
    if (ds->data_size<SHARED_BUFFER_SIZE) {
      res=acquire_next_chunk(ds);
      if (res==-1) //problem for acquiring the next chunk -> exit
	return -1;
    }

    //parsing the current data
    for (i=0;i<ds->data_size;i++) {
      //printf("%x : %x\n",i,ds->data_start[i]);

      if (ds->data_start[i]==0xffff) { //found a stop spill
	if (ds->debug)
	  printf("spill stop found at 0x%x (0x%x octects)\n",i,i*2);

	if (ignore_next) { //an corruption has been found in the beam so ignoring this one
	  ignore_next=0;
	  if (ds->debug)
	    printf("ignoring spill stop for corruption reason\n");
	} else {
	  ds->data_start=ds->data_start+i+10;
	  ds->data_size=ds->data_size-i-10;
	  return 0;
	}
      }
    }
    if (ds->debug)
      printf("found nothing\n");
    ds->data_start=ds->data_start+ds->data_size;
    ds->data_size=0;
  }
}


//parse the data beam for finding the headers and go to the data
//return 0 if everything is ok
//return -1 if no data
//return -2 if parsing problem
int parse_headers(struct dif_stream *ds) {
  unsigned short integ[2];
  int res;
  if (ds->debug)
    printf("parse_headers\n");


  //if necessary acquire a bunch of new data
  if (ds->data_size<SHARED_BUFFER_SIZE) {
    if (ds->debug)
      printf("data_size=%d\n",ds->data_size);
    res=acquire_next_chunk(ds);
    if (res==-1) {
      if (ds->debug)
	printf("cant acquire data anymore\n");
      return -1;
    }
  }


  switch(*ds->data_start) {
  case 0xfffc: //found a new spill
    integ[0]=*(ds->data_start+2);
    integ[1]=*(ds->data_start+1);
    if (ds->debug)
      printf("new spil %x\n",(*(int *)integ));
    //upgrading the data structure
    ds->current_spil=(*(int *)integ);
    ds->data_start=ds->data_start+6;
    ds->data_size=ds->data_size-6;
    //recursively calling itself for next header
    res=parse_headers(ds);
    if (res==-1)
      return -1;
    break;
  case 0xfffd: //found a new chip
    if (ds->debug)
      printf("new chip %x\n",*(((unsigned char *)(ds->data_start))+2));
    ds->nb_readout++;
    ds->current_chip=*(((unsigned char *)(ds->data_start))+2)-1;
    if (ds->current_chip>=ds->nb_chips) { // the chip number is too high, ignoring this data
      ds->nb_corrupt++;
      //printf("Big Bug : chip number is too high : %d (total=%d)\n",ds->current_chip,ds->corruptions);
      ds->data_start=ds->data_start+5;
      ds->data_size=ds->data_size-5;
      //returning an error
      return -2;
      }
    //The data is ok
    ds->data_start=ds->data_start+5;
    ds->data_size=ds->data_size-5;
    return 0;
    break;
  case 0xffff: //found a spill trailer
    integ[0]=*(ds->data_start+2);
    integ[1]=*(ds->data_start+1);
    if (ds->debug)
      printf("spil trailer %x\n",(*(int *)integ));
    ds->data_start=ds->data_start+10;
    ds->data_size=ds->data_size-10;
    //calling itself recursively for next header
    res=parse_headers(ds);
    if (res==-1)
      return -1;
    break;
  default:
    if (ds->debug)
      printf("I'm lost... moving to next spill stop\n");
    search_next_spil_stop(ds);
    //calling itself recursively for next header
    res=parse_headers(ds);
    if (res==-1)
      return -1;
  }
  //headers are finished, data is next
  return 0;
}

void display_binary(unsigned short n) {
  int i;
  printf("%4x  ",n);
  for (i=15;i>=0;i--) {
    if (i==7)
      printf(" ");
    printf("%d", (n >> i ) & 1);
  }
}

char * aggregate_data(struct dif_stream *ds,int sca,int *bc,int *gain1_en,int *gain1_hit,int *gain2_en,int *gain2_hit) {
  int i;
  char * result=malloc(1024);
  sprintf(result,"%d %d %d %d ",ds->current_chip,ds->current_spil,sca,bc[sca]);
  for (i=0;i<NB_CHANS;i++) {
    sprintf(result,"%s%c %d ",result,gain1_hit[i]>0?'T':'F',gain1_en[i]);
    sprintf(result,"%s%c %d ",result,gain2_hit[i]>0?'T':'F',gain2_en[i]);
  }
  sprintf(result,"%s\n",result);
  return result;
} 


//read the data for a chip and output it
int parse_chip(struct dif_stream *ds) {
  int i,j,k;
  unsigned short data;
  int nbcol;
  int bc[NB_SCA];
  int gain1_hit[NB_CHANS];
  int gain2_hit[NB_CHANS];
  int fbcorrupt;
  int nbhits;
  //char * result;
  
  //acquire data if necessary
  if (ds->data_size<SHARED_BUFFER_SIZE) {
    acquire_next_chunk(ds);
  }

  //analyse data
  for (i=0;i<ds->data_size;i++) {

      //searching for an error
      if (ds->data_start[i]==loss_code) {
	ds->nb_loss++;
	//printf("loss found at %d\n",i);
	return -1;
      }

      //searching a chip trailer 
      if (ds->data_start[i]==0xfffe) {
	if (ds->debug)
	  printf("data packet of size 0x%x %d words (0x%x %d bytes) \n",i,i,i*2,i*2);
	nbcol=(i-2)/129;
	if (ds->debug)
	  printf("number of columns : %d\n",nbcol);
	if (nbcol>(NB_SCA)) {
	  ds->nb_corrupt++;
	  //printf("Big Bug : number of columns is too high : %d (total=%d)\n",nbcol,ds->corruptions);
	  return -1;
	}

	//reading chip id
	ds->current_chip=ds->data_start[i-1];
	if (ds->debug)
	  printf("chip id : %d\n",ds->current_chip);
	if (ds->current_chip>=ds->nb_chips) {
	  ds->nb_corrupt++;
	  //printf("Big Bug : chip number is too high : %d (total=%d)\n",ds->current_chip,ds->corruptions);
	  return -1;
	  }

	//reading bunch crossing ids
	for (j=0;j<nbcol;j++) {
	  if (ds->debug)
	    printf("Bunch crossing: %d (0x%x)\n",ds->data_start[i-j-3],ds->data_start[i-j-3]);
	  bc[j]=ds->data_start[i-j-3];
	}
	
	fbcorrupt=0;
	//reading data
 	for (j=0;j<nbcol;j++) {	    
	  for(k=0;k<64;k++) {
	    //select good data
	    data=ds->data_start[i-3-nbcol-j*128-64-k];

	    //extract hit
	    gain1_hit[k]=data&0x1000;

	    //check if the two front bits are corrupted
	    if (data&0xc000) 
	      fbcorrupt=1;

	  }
	  for (k=0;k<64;k++) {
	    //select good data
	    data=ds->data_start[i-3-nbcol-j*128-k];
	    
	    //extract hit
	    gain2_hit[k]=data&0x1000;

	    //check if the two front bits are corrupted
	    if (data&0xc000) 
	      fbcorrupt=1;
	  }
	
	  if (fbcorrupt)
	    ds->nb_fb_corrupt++;
	  
	  nbhits=0;
	  for (k=0;k<64;k++) {
	    
	    //check if hits are coherent in the two gains
	    if (gain2_hit[k]!=gain1_hit[k]) {
	      //printf("gain conflict : g1=%d g2=%d\n",gain2_hit[k],gain1_hit[k]);
	      ds->nb_gain_corrupt++;
	    }
	    
	    //count the number of hits
	    if (gain2_hit[k]||gain1_hit[k]) {
	      ds->nb_hits++;
	      nbhits++;
	    }
	    
	  }
	  
	  //count the bcid+n events : in this implementation n=2
	  if (j>0) 
	    if (bc[j]-bc[j-1]<3)
	      ds->nb_bpn++;
	  
	  //count the plain events : in this implementation nb of hits > 10
	  if (nbhits>10) {
	    ds->nb_plain++;
	  }
	  
	  
	  //result=aggregate_data(ds,j,bc,gain1_energy,gain1_hit,gain2_energy,gain2_hit);
	  ds->nb_events++;
	  //free(result);
	}
	
	ds->data_start=ds->data_start+i+4;
	ds->data_size=ds->data_size-i-4;
	return 0;
      }
  }
  ds->nb_corrupt++;
  printf("no chip trailer found, probably a loss of packet!\n");
  return -1;
}



 
//open the data beam
 void open_data(struct dif_stream *ds) {
   if (ds->input_mode==INPUT_FILE) {
     ds->ddid=file_open(ds->filename);
     return;
   }
 }

void output_stats(struct dif_stream *ds) {
  FILE * output;
  output=fopen(ds->outputname,"w");
  fprintf(output,"%d\n",ds->nb_events);
  fprintf(output,"%d\n",ds->nb_loss);
  fprintf(output,"%d\n",ds->nb_hits);
  fprintf(output,"%d\n",ds->nb_corrupt);
  fprintf(output,"%d\n",ds->nb_gain_corrupt);
  fprintf(output,"%d\n",ds->nb_plain);
  fprintf(output,"%d\n",ds->nb_fb_corrupt);
  fprintf(output,"%d\n",ds->nb_miss_ct);
  fprintf(output,"%d\n",ds->nb_readout);
  fprintf(output,"%d\n",ds->nb_bpn);
  fclose(output);
}

int main(int argc,char **argv) {
  struct dif_stream *mydif;
  int res;

  //printf("argc=%d\n",argc);
  if (argc<3) {
    printf("usage : %s input_filename output_filename\n",argv[0]);
    exit(1);
  }

  //init the data structure 
  mydif=new_dif_stream(NB_CHIPS,NB_CHANS);
  mydif->debug=0;

  //for compatibility reason
  mydif->input_mode=INPUT_FILE;

  //get the two filename for input and output
  mydif->filename=malloc(sizeof(char)*(strlen(argv[1])+2));
  strcpy(mydif->filename,argv[1]);
  mydif->outputname=malloc(sizeof(char)*(strlen(argv[2])+2));
  strcpy(mydif->outputname,argv[2]);
  
  //opening the input device
  open_data(mydif);

  //searching for a first synchro pattern
  res=search_next_spil_stop(mydif);
  if (res==-1) { //for some reason, we cannot find a pattern
    if (mydif->debug)
      printf("cant find initial synchro\n");
    exit(1);
  }
  while (1) {
    res=parse_headers(mydif);
    if (res==-1) {
      if (mydif->debug)
	printf("cant parse headers anymore\n");
      output_stats(mydif);
      exit(1);
    }
    if (res==-2) {
      if (mydif->debug)
	printf("parse_headers is lost\n");
      continue;
    }
    res=parse_chip(mydif);
    if (res==-1) {
      if (mydif->debug)
	printf("cant parse chip\n");
      continue;
    }
  }
}
