//
// Copyright 2012-2014 Frédéric Magniette, Miguel Rubio-Roy, LLR
//

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <stdio.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <pyrame.h>

//Skiroc specific parameters

#define NB_CHANS 36
#define NB_CHIPS 1
#define NB_SCA 16

#define loss_code 0xfffa

//other parameters

#define BACKLOG 100
#define BUFFER_SIZE 1024*1024 //in 16 bits words
#define INPUT_SHMEM 1
#define INPUT_FILE 2
#define INPUT_SOCKET 3
#define OUTPUT_PIPE 4
#define OUTPUT_DEBUG 5
#define OUTPUT_SOCKET 6
#define OUTPUT_FILE 7
#define SHARED_BUFFER_SIZE 8192

struct dif_stream {
  int debug;
  int corruptions;
  //data reading
  unsigned short *buffer;
  unsigned short * data_start;
  int data_size; //in 16 bits words
  //system
  int input_mode;
  int output_mode;
  //files reading
  int ddid;
  char *filename;
  FILE * ftest;
  //shmem reading
  int difid;
  int full_semid;
  int empty_semid;
  int mutex_semid;
  int shmid;
  unsigned char * shdata;
  //socket reading
  int socket;
  char *host;
  int hport;
  //geometry
  int nb_chips;
  int nb_chans;
  //control
  unsigned int current_spil;
  unsigned int current_chip;
  //output file
  FILE *output;
  char *ofilename;
  //output socket
  unsigned short oport;
  struct netserv_client *osocket;
} dif_stream;



//this function abstract the reading mode and acquire a bunch of data
//return 0 if read is ok
//return -1 else
int acquire_next_chunk(struct dif_stream *ds) {
  int read_size;
  int total;
  unsigned char tmp[SHARED_BUFFER_SIZE];
  if (ds->debug)
    printf("acquire next chunk\n");

  //if no room left, recopy the data at the beginning of the buffer
  if (ds->data_start-ds->buffer>BUFFER_SIZE-(SHARED_BUFFER_SIZE*3)) {
    if (ds->debug)
      printf("no room left, recopy data at beginning of buffer\n");
    memcpy(ds->data_start,ds->buffer,ds->data_size);
    ds->data_start=ds->buffer;
  }

  //reading the file
  if (ds->input_mode==INPUT_FILE) {
    if (ds->debug)
      printf("mode file\n");
    //read a bunch of data 
    read_size=read(ds->ddid,ds->data_start+ds->data_size,SHARED_BUFFER_SIZE);
    if (read_size!=SHARED_BUFFER_SIZE) { 
      if (ds->debug) {
	printf("read failed %d\n",read_size);
	perror("read");
      }
      //end of file
      return -1;
    } else {
      //good reading
      ds->data_size+=SHARED_BUFFER_SIZE/2; //in words
      return 0;
    }
    //never reached
  }

  //reading the shared memory
  if (ds->input_mode==INPUT_SHMEM) {
    if (ds->debug)
      printf("mode shmem\n");
    //acquire the semaphores
    P(ds->full_semid);
    P(ds->mutex_semid);
    //copy the data
    memcpy(ds->data_start+ds->data_size,ds->shdata,SHARED_BUFFER_SIZE);
    ds->data_size+=SHARED_BUFFER_SIZE/2; //in words
    //printf("data acquired\n");
    //release the semaphores
    V(ds->mutex_semid);
    V(ds->empty_semid);
    return 0;
  }

  //read the socket
  if (ds->input_mode==INPUT_SOCKET) {
    if (ds->debug)
      printf("mode socket\n");
    total=0; //the total nb of words already read
    while(total!=SHARED_BUFFER_SIZE) {
      read_size=read(ds->socket,tmp+total,SHARED_BUFFER_SIZE-total);
      //printf("%d bytes lu\n",read_size);
      if (read_size<=0) { //the socket is closed
	if (ds->debug)
	  perror("read");
	return -1;
      }
      total=total+read_size;
    }
    //copy the data to buffer
    memcpy((unsigned char *)(ds->data_start+ds->data_size),tmp,SHARED_BUFFER_SIZE);
    ds->data_size+=SHARED_BUFFER_SIZE/2;
    return 0;
  }
  
  if (ds->debug)
    printf("impossible mode : %d\n",ds->input_mode);
  return -1;
}

void display_sample(struct dif_stream *ds) {
  int j;
  for (j=0;j<10;j++) {
    printf("%x : %x\t",j,ds->data_start[j]);
    if (j==4) printf("\n");
  }
  printf("\n");
}


//search for the 0xffff spill stop pattern in the data stream
//manage the error pattern
//return 0 if success
//return -1 if failure
int search_next_spil_stop(struct dif_stream *ds) {
  int i;
  int ignore_next=0;
  int res;
  if (ds->debug)
    printf("search_next_spil_stop\n");

  while(1) { //main loop

    //acquire data if necessary
    if (ds->data_size<SHARED_BUFFER_SIZE) {
      res=acquire_next_chunk(ds);
      if (res==-1) //problem for acquiring the next chunk -> exit
	return -1;
    }

    //parsing the current data
    for (i=0;i<ds->data_size;i++) {
      //printf("%x : %x\n",i,ds->data_start[i]);

      if (ds->data_start[i]==0xffff) { //found a stop spill
	if (ds->debug)
	  printf("spill stop found at 0x%x (0x%x octects)\n",i,i*2);

	if (ignore_next) { //an corruption has been found in the beam so ignoring this one
	  ignore_next=0;
	  if (ds->debug)
	    printf("ignoring spill stop for corruption reason\n");
	} else {
	  ds->data_start=ds->data_start+i+10;
	  ds->data_size=ds->data_size-i-10;
	  return 0;
	}
      }
    }
    if (ds->debug)
      printf("found nothing\n");
    ds->data_start=ds->data_start+ds->data_size;
    ds->data_size=0;
  }
}


//parse the data beam for finding the headers and go to the data
//return 0 if everything is ok
//return -1 if no data
//return -2 if parsing problem
int parse_headers(struct dif_stream *ds) {
  unsigned short integ[2];
  int res;
  if (ds->debug)
    printf("parse_headers\n");


  //if necessary acquire a bunch of new data
  if (ds->data_size<SHARED_BUFFER_SIZE) {
    if (ds->debug)
      printf("data_size=%d\n",ds->data_size);
    res=acquire_next_chunk(ds);
    if (res==-1) {
      if (ds->debug)
	printf("cant acquire data anymore\n");
      return -1;
    }
  }


  switch(*ds->data_start) {
  case 0xfffc: //found a new spill
    integ[0]=*(ds->data_start+2);
    integ[1]=*(ds->data_start+1);
    if (ds->debug)
      printf("new spil %x\n",(*(int *)integ));
    //upgrading the data structure
    ds->current_spil=(*(int *)integ);
    ds->data_start=ds->data_start+6;
    ds->data_size=ds->data_size-6;
    //recursively calling itself for next header
    res=parse_headers(ds);
    if (res==-1)
      return -1;
    break;
  case 0xfffd: //found a new chip
    if (ds->debug)
      printf("new chip %x\n",*(((unsigned char *)(ds->data_start))+2));
    ds->current_chip=*(((unsigned char *)(ds->data_start))+2)-1;
    if (ds->current_chip>=ds->nb_chips) { // the chip number is too high, ignoring this data
      ds->corruptions++;
      printf("Big Bug : chip number is too high : %d (total=%d)\n",ds->current_chip,ds->corruptions);
      ds->data_start=ds->data_start+5;
      ds->data_size=ds->data_size-5;
      //returning an error
      return -2;
      }
    //The data is ok
    ds->data_start=ds->data_start+5;
    ds->data_size=ds->data_size-5;
    return 0;
    break;
  case 0xffff: //found a spill trailer
    integ[0]=*(ds->data_start+2);
    integ[1]=*(ds->data_start+1);
    if (ds->debug)
      printf("spil trailer %x\n",(*(int *)integ));
    ds->data_start=ds->data_start+10;
    ds->data_size=ds->data_size-10;
    //calling itself recursively for next header
    res=parse_headers(ds);
    if (res==-1)
      return -1;
    break;
  default:
    if (ds->debug)
      printf("I'm lost... moving to next spill stop\n");
    search_next_spil_stop(ds);
    //calling itself recursively for next header
    res=parse_headers(ds);
    if (res==-1)
      return -1;
  }
  //headers are finished, data is next
  return 0;
}

void display_binary(unsigned short n) {
  int i;
  printf("%4x  ",n);
  for (i=15;i>=0;i--) {
    if (i==7)
      printf(" ");
    printf("%d", (n >> i ) & 1);
  }
}

char * aggregate_data(struct dif_stream *ds,int sca,int *bc,int *gain1_en,int *gain1_hit,int *gain2_en,int *gain2_hit) {
  int i;
  char * result=malloc(1024);
  sprintf(result,"%d %d %d %d ",ds->current_chip,ds->current_spil,sca,bc[sca]);
  for (i=0;i<NB_CHANS;i++) {
    sprintf(result,"%s%c %d ",result,gain1_hit[i]>0?'T':'F',gain1_en[i]);
    sprintf(result,"%s%c %d ",result,gain2_hit[i]>0?'T':'F',gain2_en[i]);
  }
  sprintf(result,"%s\n",result);
  return result;
} 


//read the data for a chip and output it
int parse_chip(struct dif_stream *ds) {
  int i,j,k;
  unsigned short data;
  int nbcol;
  int bc[NB_SCA];
  int gain1_energy[NB_CHANS];
  int gain2_energy[NB_CHANS];
  int gain1_hit[NB_CHANS];
  int gain2_hit[NB_CHANS];
  char * result;
  
  //acquire data if necessary
  if (ds->data_size<SHARED_BUFFER_SIZE) {
    acquire_next_chunk(ds);
  }

  //analyse data
  for (i=0;i<ds->data_size;i++) {

      //searching for an error
      if (ds->data_start[i]==loss_code) {
	printf("loss found at %d\n",i);
	return -1;
      }

      //searching a chip trailer 
      if (ds->data_start[i]==0xfffe) {
	if (ds->debug)
	  printf("data packet of size 0x%x %d words (0x%x %d bytes) \n",i,i,i*2,i*2);
	nbcol=(i-2)/129;
	if (ds->debug)
	  printf("number of columns : %d\n",nbcol);
	if (nbcol>(NB_SCA)) {
	  ds->corruptions++;
	  printf("Big Bug : number of columns is too high : %d (total=%d)\n",nbcol,ds->corruptions);
	  return -1;
	}

	//reading chip id
	ds->current_chip=ds->data_start[i-1];
	if (ds->debug)
	  printf("chip id : %d\n",ds->current_chip);
	if (ds->current_chip>=ds->nb_chips) {
	  ds->corruptions++;
	  printf("Big Bug : chip number is too high : %d (total=%d)\n",ds->current_chip,ds->corruptions);
	  return -1;
	  }

	//reading bunch crossing ids
	for (j=0;j<nbcol;j++) {
	  if (ds->debug)
	    printf("Bunch crossing: %d (0x%x)\n",ds->data_start[i-j-3],ds->data_start[i-j-3]);
	  bc[j]=ds->data_start[i-j-3];
	}
	
	
	//reading data
 	for (j=0;j<nbcol;j++) {	    
	  for(k=0;k<64;k++) {
	    //read gain1 datas
	    data=ds->data_start[i-3-nbcol-j*128-64-k];
	    gain1_energy[k]=data&0x0fff;
	    gain1_hit[k]=data&0x1000;
	  }
	  for(k=0;k<64;k++) {
	    //read gain2 datas
	    data=ds->data_start[i-3-nbcol-j*128-k];
	    gain2_energy[k]=data&0x0fff;
	    gain2_hit[k]=data&0x1000;
	  }
	  
	  result=aggregate_data(ds,j,bc,gain1_energy,gain1_hit,gain2_energy,gain2_hit);	  
	  if (ds->debug)
	    printf("result=%s",result);
	  if (ds->output_mode==OUTPUT_FILE)
	    fprintf(ds->output,"%s",result);
	  if (ds->output_mode==OUTPUT_SOCKET)
	    send2client(ds->osocket,(unsigned char *)result,strlen(result));
	  if (ds->output_mode==OUTPUT_PIPE)
	    printf("%s",result);
	  free(result);
	}
	
	ds->data_start=ds->data_start+i+4;
	ds->data_size=ds->data_size-i-4;
	return 0;
      }
  }
  printf("no chip trailer found, probably a loss of packet!\n");
  return -1;
}


//initialize a new dif_stream structure
struct dif_stream * new_dif_stream(int nb_chips,int nb_chans) {
  struct dif_stream *result;
  result=malloc(sizeof(struct dif_stream));
  memset(result,0,sizeof(struct dif_stream));
  result->buffer=malloc(BUFFER_SIZE*sizeof(unsigned short));
  result->data_start=result->buffer;
  result->nb_chips=nb_chips;
  result->nb_chans=nb_chans;
  return result;
}
 
//open the data beam
 void open_data(struct dif_stream *ds) {
   if (ds->input_mode==INPUT_FILE) {
     ds->ddid=file_open(ds->filename);
     return;
   }
   if (ds->input_mode==INPUT_SHMEM) {
     //creating semaphores and shared mem
     //ds->full_semid=create_semaphore("ipc/full",ds->difid,0,0);
     //ds->empty_semid=create_semaphore("ipc/empty",ds->difid,0,0);
     //ds->mutex_semid=create_semaphore("ipc/mutex",ds->difid,0,0);
     //ds->shdata=create_shmem("ipc/cache",ds->difid,0);
     return;
   } 
   if (ds->input_mode==INPUT_SOCKET) {
     ds->socket=open_socket(ds->host,ds->hport);
   }
 }


void usage() {
  printf("\n\n\n");
  printf("usage : skiroc -i input_mode input_parameters -o output_mode output_parameters\n");
  printf("input mode are :\n");
  printf("shmem.  Params :  difif\n");
  printf("file.   Params : filename\n");
  printf("socket. Params : host port\n");
  printf("output modes are : \n");
  printf("pipe.   Params : none\n");
  printf("debug.  Params : None\n");
  printf("file.   Params : filename\n");
  printf("socket. Params : port\n");
  exit(0);
}

int chknup(int argp,int argc) {
  int a=argp+1;
  if (a==argc) {
    printf("too few parameters\n");
    usage();
    return 0;
  } else {
    return a;
  }
}

int main(int argc,char **argv) {
  struct dif_stream *mydif;
  int res;
  int argpointer=1;

  //printf("argc=%d\n",argc);
  if (argc<2) {
    usage();
  }


  mydif=new_dif_stream(NB_CHIPS,NB_CHANS);
  mydif->debug=0;

  //check for the -i
  if (strcmp(argv[argpointer],"-i")) {
    usage();
  } else {
    argpointer=chknup(argpointer,argc);

  }
  //get the input mode
  if (!strcmp(argv[argpointer],"shmem")) {
    if (mydif->debug)
      printf("shmem input\n");
    mydif->input_mode=INPUT_SHMEM;
    argpointer=chknup(argpointer,argc);
    mydif->difid=atoi(argv[argpointer]);
    if (mydif->debug)
      printf("parameter difid=%d\n",mydif->difid);
  } else {
    if (!strcmp(argv[argpointer],"socket")) {
      if (mydif->debug)
	printf("socket input\n");
      mydif->input_mode=INPUT_SOCKET;
      argpointer=chknup(argpointer,argc);
      mydif->host=malloc(sizeof(char)*(strlen(argv[argpointer])+2));
      strcpy(mydif->host,argv[argpointer]);
      if (mydif->debug)
	printf("param host %s\n",mydif->host);
      argpointer=chknup(argpointer,argc);
      mydif->hport=atoi(argv[argpointer]);
      if (mydif->debug)
	printf("param port %d\n",mydif->hport);
    } else {
      if (!strcmp(argv[argpointer],"file")) {
	if (mydif->debug)
	  printf("file input\n");
	mydif->input_mode=INPUT_FILE;
	argpointer=chknup(argpointer,argc);
	mydif->filename=malloc(sizeof(char)*(strlen(argv[argpointer])+2));
	strcpy(mydif->filename,argv[argpointer]);
	if (mydif->debug)
	  printf("filename=%s\n",mydif->filename);
      } else {
	printf("unknown input mode : %s\n",argv[argpointer]);
	usage();
      }
    }
  }

  //check for the -o
  argpointer=chknup(argpointer,argc);
  if (strcmp(argv[argpointer],"-o"))
    usage();
  else
    argpointer=chknup(argpointer,argc);
 

  //get the output mode
  if (!strcmp(argv[argpointer],"pipe")) {
    mydif->output_mode=OUTPUT_PIPE;
    if (mydif->debug)
      printf("pipe output\n");
    mydif->debug=0;
  } else {
    if (!strcmp(argv[argpointer],"debug")) {
      mydif->output_mode=OUTPUT_DEBUG;
      if (mydif->debug)
	printf("debug output\n");
      mydif->debug=1;
    } else {
      if (!strcmp(argv[argpointer],"file")) {
      mydif->output_mode=OUTPUT_FILE;
      if (mydif->debug)
	printf("file output\n");
      argpointer=chknup(argpointer,argc);
      mydif->ofilename=malloc(sizeof(char)*(strlen(argv[argpointer])+2));
      strcpy(mydif->ofilename,argv[argpointer]);
      if (mydif->debug)
	printf("filename=%s\n",mydif->ofilename);
      } else {
	if (!strcmp(argv[argpointer],"socket")) {
	  if (mydif->debug)
	    printf("socket output\n");
	  mydif->output_mode=OUTPUT_SOCKET;
	  argpointer=chknup(argpointer,argc);
	  mydif->oport=atoi(argv[argpointer]);
	  if (mydif->debug)
	    printf("param port %d\n",mydif->oport);
	} else {
	  printf("unknown output mode : %s\n",argv[argpointer]);
	  usage();
	}
      }
    }
  }

  mydif->corruptions=0;


  //opening the ouput device
  if (mydif->output_mode==OUTPUT_FILE) {
    mydif->output=fopen(mydif->ofilename,"w");
  }
  if (mydif->output_mode==OUTPUT_SOCKET) {
    //opening the ouput sockets
    mydif->osocket=first_client(start_monoport_netserv(mydif->oport,10));
  }

  //opening the input device
  open_data(mydif);
  //searching for a first synchro pattern
  res=search_next_spil_stop(mydif);
  if (res==-1) { //for some reason, we cannot find a pattern
    if (mydif->debug)
      printf("cant find initial synchro\n");
    exit(1);
  }
  while (1) {
    res=parse_headers(mydif);
    if (res==-1) {
      if (mydif->debug)
	printf("cant parse headers anymore\n");
      fclose(mydif->output);
      exit(1);
    }
    if (res==-2) {
      if (mydif->debug)
	printf("parse_headers is lost\n");
      continue;
    }
    res=parse_chip(mydif);
    if (res==-1) {
      if (mydif->debug)
	printf("cant parse chip\n");
      continue;
    }
  }
}

